@@ -27,6 +27,14 @@ def to_unicode(s):
     else:
         return s
 
+def get_ns(ns):
+    out_ns = to_unicode(libxml2mod.xmlNodeGetContent(ns))
+    # Detect "" and produce None as the empty namespace.
+    if out_ns:
+        return out_ns
+    else:
+        return None
+
 def _get_prefix_and_localName(name):
     t = name.split(":")
     if len(t) == 1:
@@ -37,6 +45,35 @@ def _get_prefix_and_localName(name):
         # NOTE: Should raise an exception.
         return None, None
 
+def _find_namespace(node, ns, prefix):
+    new_ns = None
+    current = libxml2mod.xmlNodeGetNsDefs(node)
+    while current is not None:
+        if _check_namespace(current, ns, prefix):
+            new_ns = current
+            break
+        current = libxml2mod.next(current)
+    if new_ns is None:
+        node_ns = libxml2mod.xmlNodeGetNs(node)
+        if node_ns is not None and _check_namespace(node_ns, ns, prefix):
+            new_ns = node_ns
+    return new_ns
+
+def _check_namespace(current, ns, prefix):
+    current_ns = libxml2mod.xmlNodeGetContent(current)
+    current_prefix = libxml2mod.name(current)
+    if ns == current_ns and prefix == current_prefix:
+        return 1
+    else:
+        return 0
+
+def _make_namespace(node, ns, prefix, set_default=0):
+    if prefix is not None or set_default:
+        new_ns = libxml2mod.xmlNewNs(node, ns, prefix)
+    else:
+        new_ns = None
+    return new_ns
+
 _nodeTypes = {
     "attribute" : xml.dom.Node.ATTRIBUTE_NODE,
     "comment" : xml.dom.Node.COMMENT_NODE,
@@ -80,7 +117,7 @@ def Node_attributes(node):
     while node is not None:
         ns = libxml2mod.xmlNodeGetNs(node)
         if ns is not None:
-            attributes[(libxml2mod.xmlNodeGetContent(ns), libxml2mod.name(node))] = node
+            attributes[(get_ns(ns), libxml2mod.name(node))] = node
         else:
             attributes[(None, libxml2mod.name(node))] = node
         node = libxml2mod.next(node)
@@ -89,7 +126,7 @@ def Node_attributes(node):
 def Node_namespaceURI(node):
     ns = libxml2mod.xmlNodeGetNs(node)
     if ns is not None:
-        return to_unicode(libxml2mod.xmlNodeGetContent(ns))
+        return get_ns(ns)
     else:
         return None
 
@@ -172,25 +209,22 @@ def Node_getAttributeNode(node, name):
     return Node_attributes(node)[(None, name)]
 
 def Node_setAttributeNS(node, ns, name, value):
-    # NOTE: Need to convert from Unicode.
     ns, name, value = map(from_unicode, [ns, name, value])
-
     prefix, localName = _get_prefix_and_localName(name)
-
-    # NOTE: Might need to be xmlSetNsProp.
-    if ns is not None and ns == libxml2mod.xmlNodeGetContent(libxml2mod.xmlNodeGetNs(node)):
-        libxml2mod.xmlNewNsProp(node, libxml2mod.xmlNodeGetNs(node), localName, value)
-    elif prefix is not None:
-        new_ns = libxml2mod.xmlNewNs(node, ns, prefix)
-        libxml2mod.xmlNewNsProp(node, new_ns, localName, value)
+    if ns is not None:
+        if prefix == "xmlns" and ns == xml.dom.XMLNS_NAMESPACE and _find_namespace(node, value, localName):
+            return
+        new_ns = _find_namespace(node, ns, prefix)
+        if new_ns is None:
+            new_ns = _make_namespace(node, ns, prefix, set_default=0)
+        libxml2mod.xmlSetNsProp(node, new_ns, localName, value)
     else:
         # NOTE: Needs verifying: what should happen to the namespace?
         # NOTE: This also catches the case where None is the element's
         # NOTE: namespace and is also used for the attribute.
-        libxml2mod.xmlNewNsProp(node, None, localName, value)
+        libxml2mod.xmlSetNsProp(node, None, localName, value)
 
 def Node_setAttribute(node, name, value):
-    # NOTE: Need to convert from Unicode.
     name, value = map(from_unicode, [name, value])
 
     libxml2mod.xmlSetProp(node, name, value)
@@ -212,54 +246,60 @@ def Node_removeAttribute(node, name):
     libxml2mod.xmlUnsetProp(node, name)
 
 def Node_createElementNS(node, ns, name):
-    # NOTE: Need to convert from Unicode.
     ns, name = map(from_unicode, [ns, name])
 
     prefix, localName = _get_prefix_and_localName(name)
     new_node = libxml2mod.xmlNewNode(localName)
-    # NOTE: Does it make sense to set the namespace if it is empty?
+
+    # If the namespace is not empty, set the declaration.
     if ns is not None:
-        new_ns = libxml2mod.xmlNewNs(new_node, ns, prefix)
+        new_ns = _find_namespace(new_node, ns, prefix)
+        if new_ns is None:
+            new_ns = _make_namespace(new_node, ns, prefix, set_default=1)
         libxml2mod.xmlSetNs(new_node, new_ns)
+    # If the namespace is empty, set a "null" declaration.
+    elif prefix is not None:
+        new_ns = _find_namespace(new_node, "", prefix)
+        if new_ns is None:
+            new_ns = _make_namespace(new_node, "", prefix)
+        libxml2mod.xmlSetNs(new_node, new_ns)
+    else:
+        libxml2mod.xmlSetNs(new_node, None)
+        Node_setAttribute(new_node, "xmlns", "")
     return new_node
 
 def Node_createElement(node, name):
-    # NOTE: Need to convert from Unicode.
     name = from_unicode(name)
 
     new_node = libxml2mod.xmlNewNode(name)
     return new_node
 
 def Node_createAttributeNS(node, ns, name):
-
-    # NOTE: Need to convert from Unicode.
     ns, name = map(from_unicode, [ns, name])
 
     prefix, localName = _get_prefix_and_localName(name)
     # NOTE: Does it make sense to set the namespace if it is empty?
     if ns is not None:
-        new_ns = libxml2mod.xmlNewNs(node, ns, prefix)
+        new_ns = _find_namespace(node, ns, prefix)
+        if new_ns is None:
+            new_ns = _make_namespace(node, ns, prefix, set_default=0)
     else:
         new_ns = None
     new_node = libxml2mod.xmlNewNsProp(node, new_ns, localName, None)
     return new_node
 
 def Node_createAttribute(node, name):
-
-    # NOTE: Need to convert from Unicode.
     name = from_unicode(name)
 
     # NOTE: xmlNewProp does not seem to work.
     return Node_createAttributeNS(node, None, name)
 
 def Node_createTextNode(node, value):
-    # NOTE: Need to convert from Unicode.
     value = from_unicode(value)
 
     return libxml2mod.xmlNewText(value)
 
 def Node_createComment(node, value):
-    # NOTE: Need to convert from Unicode.
     value = from_unicode(value)
 
     return libxml2mod.xmlNewComment(value)
@@ -321,7 +361,6 @@ def Node_importNode_DOM(node, other, deep):
     raise ValueError, "Node type '%s' (%d) not supported." % (_reverseNodeTypes[other.nodeType], other.nodeType)
 
 def Node_xpath(node, expr, variables=None, namespaces=None):
-    # NOTE: Need to convert from Unicode.
     expr = from_unicode(expr)
 
     context = libxml2mod.xmlXPathNewContext(Node_ownerDocument(node))
@@ -333,7 +372,6 @@ def Node_xpath(node, expr, variables=None, namespaces=None):
         libxml2mod.xmlXPathRegisterNs(context, prefix, ns)
     # NOTE: No such functions are exposed in current versions of libxml2.
     #for (prefix, ns), value in (variables or {}).items():
-    #    # NOTE: Need to convert from Unicode.
     #    value = from_unicode(value)
     #    libxml2mod.xmlXPathRegisterVariableNS(context, prefix, ns, value)
     result = libxml2mod.xmlXPathEval(expr, context)
@@ -346,6 +384,7 @@ def createDocument(namespaceURI, localName, doctype):
     # NOTE: Fixed to use version 1.0 only.
     d = libxml2mod.xmlNewDoc("1.0")
     if localName is not None:
+        # NOTE: Verify that this is always what should occur.
         root = Node_createElementNS(d, namespaceURI, localName)
         Node_appendChild(d, root)
     if doctype is not None:
